import socket
import string
import gc
import sys
import json
from collections.abc import Iterable
import traceback

import torch
import accelerate
from threading import Event
from transformers import AutoModelForCausalLM, AutoTokenizer, LlamaTokenizer, TextIteratorStreamer
from transformers.generation.logits_process import (
    LogitsProcessorList,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,
)


#from fastchat.conversation import get_conv_template, SeparatorStyle
#from fastchat.model.model_adapter import load_model, get_conversation_template
#from fastchat.model.chatglm_model import chatglm_generate_stream
#from fastchat.model.falcon_model import falcon_generate_stream
#from fastchat.modules.gptq import GptqConfig
#from fastchat.utils import is_partial_stop


def prepare_logits_processor(
    temperature: float, repetition_penalty: float, top_p: float, top_k: int
) -> LogitsProcessorList:
    processor_list = LogitsProcessorList()
    # TemperatureLogitsWarper doesn't accept 0.0, 1.0 makes it a no-op so we skip two cases.
    if temperature >= 1e-5 and temperature != 1.0:
        processor_list.append(TemperatureLogitsWarper(temperature))
    if repetition_penalty > 1.0:
        processor_list.append(RepetitionPenaltyLogitsProcessor(repetition_penalty))
    if 1e-8 <= top_p < 1.0:
        processor_list.append(TopPLogitsWarper(top_p))
    if top_k > 0:
        processor_list.append(TopKLogitsWarper(top_k))
    return processor_list


def infer(model, tokenizer, prompt, temperature=0.2, top_p=1.0, context_len=2048, max_new_tokens=50, stop_str=None, echo=False, stop_token_ids = []):
    device = torch.cuda.current_device()
    input_ids = tokenizer(prompt).input_ids
    output_ids = list(input_ids)
    past_key_values = out = None
    len_prompt = len(prompt)
    repetition_penalty = 1.0
    top_k = 50
    stop_token_ids.append(tokenizer.eos_token_id)
    stream_interval = 2
    max_src_len = context_len - max_new_tokens - 8
    input_ids = input_ids[-max_src_len:]
    input_echo_len = len(input_ids)
    
    logits_processor = prepare_logits_processor(
        temperature, repetition_penalty, top_p, top_k
    )

    print(f'max new tokens {max_new_tokens}')
    for i in range(max_new_tokens):
        print(i)
        if i == 0:
            out = model(torch.as_tensor([input_ids], device=device), use_cache=True)
        else:
           try:
               out = model(
                input_ids=torch.as_tensor([[token]], device=device),
                use_cache=True,
                past_key_values=past_key_values,
               )
           except Exception as e:
               print(str(e))
               traceback.print_exc()
        logits = out.logits
        past_key_values = out.past_key_values
        last_token_logits = logits[0, -1, :]
        
        probs = torch.softmax(last_token_logits, dim=-1)
        token = int(torch.multinomial(probs, num_samples=1))

        output_ids.append(token)

        if token in stop_token_ids:
            stopped = True
        else:
            stopped = False

        if i % stream_interval == 0 or i == max_new_tokens - 1 or stopped:
            if echo:
                tmp_output_ids = output_ids
                rfind_start = len_prompt
            else:
                tmp_output_ids = output_ids[input_echo_len:]
                rfind_start = 0

            output = tokenizer.decode(
                tmp_output_ids,
                skip_special_tokens=True,
                spaces_between_special_tokens=False,
            )

            partially_stopped = False
            if stop_str:
                if isinstance(stop_str, str):
                    pos = output.rfind(stop_str, rfind_start)
                    if pos != -1:
                        output = output[:pos]
                        stopped = True
                    else:
                        partially_stopped = is_partial_stop(output, stop_str)

                elif isinstance(stop_str, Iterable):
                    for each_stop in stop_str:
                        pos = output.rfind(each_stop, rfind_start)
                        if pos != -1:
                            output = output[:pos]
                            stopped = True
                            break
                        else:
                            partially_stopped = is_partial_stop(output, each_stop)
                            if partially_stopped:
                                break
                else:
                    print('Value Error')
                    raise ValueError("Invalid stop field type.")

            # prevent yielding partial stop sequence
            if not partially_stopped:
                print(' stream')
                yield {
                    "text": output,
                    "usage": {
                        "prompt_tokens": input_echo_len,
                        "completion_tokens": i,
                        "total_tokens": input_echo_len + i,
                    },
                    "finish_reason": None,
                }

        if stopped:
            break

# finish stream event, which contains finish reason
    if i == max_new_tokens - 1:
        finish_reason = "length"
    elif stopped:
        finish_reason = "stop"
    else:
        finish_reason = None
    print(f'finish {finish_reason}')
    yield {
        "text": output,
        "usage": {
            "prompt_tokens": input_echo_len,
            "completion_tokens": i,
            "total_tokens": input_echo_len + i,
        },
        "finish_reason": finish_reason,
    }

    # clean
    del past_key_values, out
    gc.collect()
    torch.cuda.empty_cache()


    
def submit(message, model=None, tokenizer=None, pipeline=None, stop_event=None,conn=None, stop_str = None):
    message_j = json.loads(message)
    temp = 0.7
    if 'temp' in message_j.keys():
        temp = message_j['temp']
    top_p = 1.0
    if 'top_p' in message_j.keys():
        top_p = message_j['top_p']
    max_tokens = 100
    if 'max_tokens' in message_j.keys():
        max_tokens = message_j['max_tokens']
    user_prompt = 'User'
    if 'user' in message_j.keys():
        user_prompt = message_j['user']
    asst_prompt = 'Assistant'
    if 'asst' in message_j.keys():
        asst_prompt = message_j['asst']
    eos = '<|endoftext|>'
    if 'eos' in message_j.keys():
        eos = message_j['eos']
    prompt = message_j['prompt']
    
    print(f'\n temp {temp}, top_p: {top_p}, max_tokens={max_tokens}\n')

    result_stream = infer(model, tokenizer, prompt, temp, top_p, max_new_tokens=max_tokens, stop_token_ids=[eos])
    generated_text = ''
    for chunk in result_stream:
        new_text = chunk['text']
        print(new_text)
        test_text = generated_text + new_text
        idx1 = test_text.find('<|endoftext|>')
        idx2 = test_text.find(user_prompt)
        idx3 = -1
        if stop_str is not None and isinstance(stop_str, Iterable):
          for stop in stop_str:
            idx3 = max(idx3, test_text.find(stop))
        if stop_event.is_set(): # flush generate
          continue
        if idx1>= 0 or idx2>=0 or idx3 >= 0:
          stop_event.set()
          idx = min(max(idx1, 0), max(idx2,0), max(idx3,0))
          if idx > 0:
            conn.send(bytearray((new_text[:idx]).encode('utf8')))
            print(f'set stop event in submit {idx1}, {idx2}, {idx3}')
        else:
          generated_text = test_text
          conn.send(bytearray((new_text).encode('utf8')))  # send data to the client
    return generated_text



def server(model=None, tokenizer=None, pipeline=None, stop_str=None):
    # either model/tokenizer OR pipeline.
    # stop_str is a list of stop_strings
    # get the hostname
    host = socket.gethostname()
    host = ''
    port = 5004  # initiate port no above 1024
    stop_event = Event()
    print(f"starting server {stop_str}")
    while True:
      try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM)  as server_socket: # get instance
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind((host, port))  # bind host address and port together
            server_socket.settimeout(3)

            while True:
                try:
                    conn = None
                    server_socket.listen(10)
                    #print("waiting...")
                    try:
                      conn, address = server_socket.accept()  # accept new connection
                    except Exception:
                      pass
                    if conn is None: continue
                    print("Connection from: " + str(address))
                    query_string = ''
                    while True:
                      s = conn.recv(1024)
                      if s is None or not s:
                        break
                      if (len(s) > 5 and s[-3:] == b'xff' and s[-6:-3] == b'x00'):
                        s = s[:-6]
                        query_string += s.decode('utf-8')
                        break
                      else:
                        query_string += s.decode('utf-8')
                    print("got string:", query_string)
                    if query_string is not None and len(query_string) > 0:
                        message_j = json.loads(query_string)
                        submit(query_string, model=model, tokenizer=tokenizer, pipeline=pipeline, stop_event=stop_event, conn=conn, stop_str=stop_str)
                        print('sending termination')
                        conn.sendall(b'x00xff')
                    if conn is not None: conn.close()
                except TimeoutError:
                  #traceback.print_exc()
                  if conn is not None: conn.close()
                except KeyboardInterrupt:
                  print("idle loop interrrupt")
                  traceback.print_exc()
                  sys.exit(0)
        print("resetting socket")
        server_socket.close()
      except BrokenPipeError:
        pass
      except KeyboardInterrupt:
        sys.exit(0)
      except Exception:
        traceback.print_exc()


























if __name__ == '__main__':
    model_name = "JosephusCheung/Guanaco"

    print('devices', torch.cuda.device_count(), torch.cuda.current_device())
    print(f"Starting to load the model {model_name} into memory")
    
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForCausalLM.from_pretrained(
        model_name,
        device_map="auto",
    )
    model.tie_weights()

    prev_text_len = 0
    while True:
        prompt = input('?')
        for value in infer(model, tokenizer, device, prompt):
            text=value['text']
            text = "".join(filter(lambda x: x in string.printable, text))
            print(value['text'][prev_text_len:])
            prev_text_len = len(value['text'])
            if value['finish_reason'] is not None:
                print(value['finish_reason'])
                break



from promptrix.promptrixTypes import Message, PromptFunctions, PromptMemory, Tokenizer
from alphawave.alphawaveTypes import PromptResponse, Validation, PromptResponseValidator
from alphawave.Response import Response
import json
import ast
import traceback
import re
from pylint.lint import Run
from pylint.reporters.text import TextReporter
from io import StringIO
import subprocess

class WritableObject(object):
    """"output stream for pylint"""
    def __init__(self):
        self.content = []
    def write(self, st):
        self.content.append(st)
    def read(self):
        return self.content

class PythonResponseValidator(PromptResponseValidator):
    def __init__(self):
        pass

    def get_python(self, text):
        # change - there may be imports before class
    
        key = "```"
        key_len = len(key)
        start = text.find(key)
        if start < 0:
            #print ('****** python extract cant find starting key')
            return text, text
        text = text[start+key_len:].lstrip(' \n,')
        #print(f'****** python extract starting on {text[:10]}')
        if text.startswith('python'):
            #print(f'****** python extract stripping starting "python"')
            text = text[6:]
            text = text.lstrip('\n')
            #print(f'****** python extract stripped starting "python" {text[:24]}')
        
        # look for class name
        lines = text.split('\n')
        class_name = None
        for line in lines:
            #print(f'   {line}')
            if line.lstrip(' \n').startswith('class'):
                class_name = (line[6:].strip()).replace(":", '')
                if '(' in class_name:
                    class_name = class_name[:class_name.find('(')]
                print(f'     found class name {class_name}')
                break
        end = text.find("```")
        if end < 0:
            #print('****** python extract cant find end key')
            return class_name, text
        return class_name, text[:end].rstrip()

    
    def validate_response(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer , response: PromptResponse, remaining_attempts: int = 0) -> Validation:
        message = response['message']
        raw_text = message if isinstance(message, str) else message.get('content', '')
        class_name, code = self.get_python(raw_text)
        if len(code) == 0:
            #print(f'***** PythonResponseValidator failure len(parsed) == 0')
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': 'No python code found'
            }

        # Validate the response against pylint
        try:
            #print(f'***** PythonResponseValidator code \n{code}\n')
            with open(class_name+'.py', 'w') as cd:
                cd.write(code)
            pylint_output = WritableObject()
            pylint_output.content=[]

            pylint_out = subprocess.run(['pylint', '--errors-only', class_name+'.py'], stdout=subprocess.PIPE)
            #Run(['--errors-only', class_name+'.py'], reporter=TextReporter(pylint_output))
            error_msgs = ''
            for l in (pylint_out.stdout).decode('utf-8').split('\n'):
                if len(l) > 4 and not l.startswith('***'):
                    error_msgs += l
            if len(error_msgs) > 0:
                #print(f'***** PythonResponseValidator code failed \n{error_msgs}\n')
                return {
                    'type': 'Validation',
                    'valid': False,
                    'feedback': error_msgs
                }
            else:
                #print(f'***** PythonResponseValidator code valid')
                return {
                    'type': 'Validation',
                    'valid': True,
                    'value': code
                }
        except Exception as e:
            #print(f'***** PythonResponseValidator exception {str(e)}')
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': str(e)
            }
            



from promptrix.promptrixTypes import PromptMemory
from promptrix.VolatileMemory import VolatileMemory

class MemoryFork(PromptMemory):
    def __init__(self, memory):
        self._fork = VolatileMemory()
        self._memory = memory

    def has(self, key):
        return self._fork.has(key) or self._memory.has(key)

    def get(self, key):
        if self._fork.has(key):
            value = self._fork.get(key)
            if value is not None and isinstance(value, dict):
                return json.loads(json.dumps(value))
            else:
                return value
        else:
            value = self._memory.get(key)
            if value is not None and isinstance(value, dict):
                return json.loads(json.dumps(value))
            else:
                return value

    def set(self, key, value):
        if value is not None and isinstance(value, dict):
            clone = json.loads(json.dumps(value))
            self._fork.set(key, clone)
        else:
            self._fork.set(key, value)

    def delete(self, key):
        if self._fork.has(key):
            self._fork.delete(key)

    def clear(self):
        self._fork.clear()



import torch

from typing import Any, Dict
from transformers import AutoModelForCausalLM, AutoTokenizer


class EndpointHandler:
    def __init__(self, path=""):
        # load model and tokenizer from path
        self.tokenizer = AutoTokenizer.from_pretrained(path)
        self.model = AutoModelForCausalLM.from_pretrained(
            path, device_map="auto", torch_dtype=torch.float16, trust_remote_code=True
        )
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def __call__(self, data: Dict[str, Any]) -> Dict[str, str]:
        # process input
        inputs = data.pop("inputs", data)
        parameters = data.pop("parameters", None)

        # preprocess
        inputs = self.tokenizer(inputs, return_tensors="pt").to(self.device)

        # pass inputs with all kwargs in data
        if parameters is not None:
            outputs = self.model.generate(**inputs, **parameters)
        else:
            outputs = self.model.generate(**inputs)

        # postprocess the prediction
        prediction = self.tokenizer.decode(outputs[0], skip_special_tokens=True)

        return [{"generated_text": prediction}]


from pyee import AsyncIOEventEmitter
from jsonschema import validate, ValidationError
import toml
from toml import TomlDecodeError
from cerberus import Validator
from typing import Any, Dict, Optional, Union
import json
from promptrix.promptrixTypes import PromptMemory, PromptFunctions, Tokenizer
from dataclasses import dataclass, asdict
import traceback

@dataclass
class CommandSchema:
    def __init__(self, type: str, title: str='', description: str='', returns: Optional[str] = None):
        self.type = type
        self.title = title
        self.description = description
        self.returns = returns

class SchemaBasedCommand(AsyncIOEventEmitter):
    def __init__(self, schema: CommandSchema, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__()
        self._schema = schema
        if self._schema is not None and type(self._schema) != dict:
            self._schema = self._schema.__dict__
        self._title = title
        self._description = description

    @property
    def description(self) -> Optional[str]:
        return self._description or self._schema['description']

    ### needs to be adapted for TOML
    @property
    def inputs(self) -> Optional[str]:
        if self._schema['properties']:
            properties = self._schema['properties'] or {}
            inputs = [f'"{key}":"<{property.get("description", property.get("type", "any"))}>"' for key, property in properties.items()]
            return ",".join(inputs)
        else:
            return None
    @property
    def output(self) -> Optional[str]:
        return self._schema['returns']

    @property
    def schema(self) -> CommandSchema:
        return self._schema

    @property
    def title(self) -> str:
        return self._title or self._schema['title']

    def execute(self, inputs: Dict[str, Any], memory: 'PromptMemory', functions: 'PromptFunctions', tokenizer: 'Tokenizer') -> Any:
        raise NotImplementedError

    def validate(self, inputs: Dict[str, Any], memory: 'PromptMemory', functions: 'PromptFunctions', tokenizer: 'Tokenizer', syntax: str ='JSON') -> 'Validation':
        #print(f'***** SchemaBasedCommand validate inputs {inputs}\nschema\n{self._schema}')
        if self._schema is None:
            return {
                'type': 'Validation',
                'valid': True,
                'value': cleaned
            }
        if syntax == 'JSON':
            try:
                cleaned = self.clean_inputs(inputs)
                #print(f'***** SchemaBasedCommand validate assuming JSON schema:\n{self._schema}\ninputs:\n{inputs}\n')
                validate(cleaned, self._schema)
                #print(f'***** SchemaBasedCommand validate passed\n')
                return {
                    'type': 'Validation',
                    'valid': True,
                    'value': cleaned
                }
            except ValidationError as e:
                #print(f'***** SchemaBasedCommand validate fail {str(e)}\n')
                errors = f'"{e.path[0] if e.path else "inputs"}": {e.message}'
                message = errors
                if 'is a required property' in message:
                    return {
                        'type': 'Validation',
                        'valid': False,
                        'feedback': f"The ['command']['inputs'] field has errors:\n{message}\nRevise to use this format: {self._schema}"
                    }
                return {
                    'type': 'Validation',
                    'valid': False,
                    'feedback': f"The ['command']['inputs'] field has errors:\n{message}\n\nTry again."
                }
        else: ## not JSON, assume TOML
            #print(f'***** SchemaBasedCommand validate assuming TOML inputs type {type(inputs)}')
            try:
                if type(inputs) != dict:
                    #print(f'***** SchemaBasedCommand validate TOML inputs not dict')
                    inputs_as_dict = None
                    try:
                        inputs_as_dict = toml.loads(inputs)
                        #print(f'***** SchemaBasedCommand validate TOML inputs toml.loads success')
                    except:
                        try:
                            inputs_as_dict = json.loads(inputs)
                            #print(f'***** SchemaBasedCommand validate TOML json.loads success')
                        except:
                            pass
                    if inputs_as_dict is None:
                        #print(f'***** SchemaBasedCommand validate TOML inputs convert to dict fail')
                        return {
                            'type': 'Validation',
                            'valid': False,
                            'feedback': f"The ['command']['inputs'] field has errors, use this format: {self._schema}"
                        }

                else:
                    inputs_as_dict = inputs

                cleaned = self.clean_inputs(inputs_as_dict)
                #print(f'***** SchemaBasedCommand validate TOML inputs cleaned\n{cleaned}')
                validate(cleaned, self._schema)
                return {
                    'type': 'Validation',
                    'valid': True,
                    'value': cleaned
                }
            except Exception as e:
                return {
                    'type': 'Validation',
                    'valid': False,
                    'feedback': f"The ['command']['inputs'] field has errors: {str(e)}. use this format: {self._schema}"
                }
                
            
            
    def clean_inputs(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        #print(f'***** SchemaBasedCommand clean_inputs inputs\n{inputs}\n')
        cleaned = {}
        properties = self._schema['properties']
        if not properties:
            return cleaned

        for key, property in properties.items():
            type = property.get('type', 'any')
            value = inputs.get(key)

            if value is None or (isinstance(value, str) and value.startswith("<") and value.endswith(">")):
                continue

            if not isinstance(value, str):
                value = json.dumps(value) if isinstance(value, dict) else str(value)

            try:
                if type == 'string':
                    cleaned[key] = str(value)
                elif type == 'number':
                    cleaned[key] = float(value)
                elif type == 'boolean':
                    cleaned[key] = bool(value)
                elif type in ['array', 'object']:
                    cleaned[key] = json.loads(value)
            except Exception:
                pass

        return cleaned

    def one_shot(self, syntax):
        content=''
        if syntax == 'JSON':
            content += '\t\tformat: {'+f'"command":'+f'"{self.title}", "inputs":'+'{'
            args = ''
            for arg in self.inputs.split(','):
                arg_key_dscp = arg.split(':')
                key = arg_key_dscp[0].strip().strip('"\'') # strip key down to bare alpha key (':' elims type info)
                content += f'"{key}": "<{arg_key_dscp[1].strip()}>",'
                content = content[:-1]+'}}\n' # strip final comma 
            #print(f'***** AgentCommandSection one-shot prompt: {content}\n')
            return content
        else:
            content += f'\t\tformat:\n[RESPONSE]\ncommand="{self.title}"\n'
            args = ''
            for arg in self.inputs.split(','):
                arg_key_dscp = arg.split(':')
                key = 'inputs.'+arg_key_dscp[0].strip().strip('"\'') # strip key down to bare alpha key (':' elims type info)
                content += f'{key}="<{arg_key_dscp[1].strip()}>"\n'
                content +='[STOP]\n'
            #print(f'***** AgentCommandSection one-shot TOML: {content}\n')
            return content



import argparse, json, os, sys
import asyncio
import traceback
from datetime import datetime, date
import tkinter as tk
import tkinter.font
from tkinter import ttk
from promptrix.VolatileMemory import VolatileMemory
from promptrix.FunctionRegistry import FunctionRegistry
from promptrix.GPT3Tokenizer import GPT3Tokenizer
from promptrix.Prompt import Prompt
from promptrix.SystemMessage import SystemMessage
from promptrix.UserMessage import UserMessage
from promptrix.AssistantMessage import AssistantMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.promptrixTypes import Message, PromptFunctions, PromptMemory, Tokenizer
from promptrix.GPT3Tokenizer import GPT3Tokenizer
from promptrix.VolatileMemory import VolatileMemory
from alphawave.alphawaveTypes import PromptCompletionOptions, PromptResponse, PromptResponseValidator, Validation
from alphawave.DefaultResponseValidator import DefaultResponseValidator
from alphawave.JSONResponseValidator import JSONResponseValidator
from alphawave.RepairTestClient import TestClient as TestClient
from alphawave.AlphaWave import AlphaWave
from alphawave.OpenAIClient import OpenAIClient
from alphawave.OSClient import OSClient
import alphawave_pyexts.LLMClient as llm
from alphawave_agents.Agent import Agent, AgentOptions
from alphawave_agents.AskCommand import AskCommand
from alphawave_agents.MathCommand import MathCommand
from alphawave_agents.FinalAnswerCommand import FinalAnswerCommand
from alphawave_pyexts.SearchCommand import SearchCommand
import embedding_wv2 as em
import usermodel_class as um

parser = argparse.ArgumentParser()

parser.add_argument('model', type=str, default='wizardLM', choices=llm.get_available_models().append('gpt-4'),help='prompt template model')
args = parser.parse_args()
if args.model is not None:
    model = args.model
else:
    model = 'vicuna_v1.1'

prompt_text = ''

if model.startswith('gpt'):
    client = OpenAIClient(apiKey=os.getenv('OPENAI_API_KEY'), logRequests=True)
else:
    #create OS client
    client = OSClient(apiKey=None)#, logRequests=True)

today_prime = "Todays date is "+date.today().strftime("%b-%d-%Y")+'. The current time is '+datetime.now().strftime('%M minutes after %H hours local time')+'.\n\n'

try:
    em.open()
except Exception as e:
  print(f'Failed to load memory.{str(e)}')

    
def make_agent():
    global client, model, prompt_text, memory
    agent_options = AgentOptions(
        client = client,
        prompt=[prompt_text],
        prompt_options=PromptCompletionOptions(
            completion_type = 'chat',
            model = model,
            temperature = 0.1,
            max_input_tokens = 3600,
            max_tokens = 800,
        ),
        step_delay=0,
        max_steps=30,
        max_repair_attempts=2,
        syntax='JSON',
        tokenizer=GPT3Tokenizer(),
        #logRepairs=True
    )

    # Create an agent
    agent = Agent(options = agent_options)
    agent.addCommand(AskCommand())
    agent.addCommand(FinalAnswerCommand())
    agent.addCommand(MathCommand())
    agent.addCommand(SearchCommand(OpenAIClient(apiKey=os.getenv('OPENAI_API_KEY')), model='gpt-3.5-turbo'))
    memory = agent._options['memory']
    return agent

agent = make_agent()

print(agent._options['prompt_options'].temperature)

def run_query(query):
    global agent
    try:
        response = asyncio.run(agent.completeTask(query))
        if type(response) == dict and 'type' in response and response['type'] == 'TaskResponse':
            if response['status'] == 'success' or response['status'] == 'input_needed':
                input_area.insert(tk.END, response['message'])
                #input_area.insert(tk.END, '\n')
    except Exception:
        traceback.print_exc()
        
def show_prompt():
    pass

def set_model():
    pass

def set_max_tokens(maxt):
    agent._options['prompt_options'].max_tokens=maxt

def set_temperature(temp):
    agent._options['prompt_options'].temperature=temp

def set_top_p(top):
    agent._options['prompt_options'].top_p=top

def set_max_tokens(maxt):
    agent._options['prompt_options'].max_tokens=maxt

def set(text):
    global temperature, top_p, max_tknns
    try:
        if 'temp' in text:
            idx = text.find('=')
            if idx > 0:
                temperature = float(text[idx+1:].strip())
        if 'top_p' in text:
            idx = text.find('=')
            if idx > 0:
                top_p = float(text[idx+1:].strip())
        if 'max_tkns' in text:
            idx = text.find('=')
            if idx > 0:
                max_tkns = int(text[idx+1:].strip())
    except Exception as e:
        print ("parse failed", e)

PREV_LEN = 0
def submit():
    global PREV_LEN
    input_text = input_area.get("1.0", tk.END)
    if len(input_text) > PREV_LEN:
        new_text = input_text[PREV_LEN:].strip()
    else:
        new_text = input_text.strip()
    #input_area.insert(tk.END, '\n')


    set_max_tokens(int(max_tkns.get()))
    set_temperature(float(temperature.get()))
    set_top_p(float(top_p.get()))
    response = run_query(new_text)
    input_area.insert(tk.END, '\n')
    PREV_LEN=len(input_area.get("1.0", tk.END))-1
    return(input, response)

def setPrompt():
    global prompt_text, agent
    prompt_text = input_area.get("1.0", tk.END)
    prompt_text = prompt_text.strip()
    if prompt_text.startswith("NoPrompt"):
        prompt_text = ''
    if prompt_text.startswith("Helpfull"):
        prompt_text = 'Respond as a helpful, friendly, information assistant. Say "I don\'t know" if you have no information"'
    if prompt_text.startswith("Bhagaven"):
        prompt_text = 'Respond as a compassionate, empathetic, self-realized follower of Ramana Maharshi'
    if prompt_text.startswith("ACT"):
        prompt_text = 'Respond as a compassionate, empathetic, counselor familiar with Acceptance Commitment Theray'
    if prompt_text.startswith("Flirt"):
        prompt_text = 'Respond as a friendly, chatty, flirty young woman named Samantha. You are 21 years old.'
    agent = make_agent()

def clear():
    global memory, PREV_POS
    PREV_POS="1.0"
    input_area.delete("1.0", tk.END)
    memory.set('history', [])
    
root = tk.Tk()

root.title(args.model)
root.geometry("1440x1280")
#root.config(cursor="watch")
root.rowconfigure(0, weight=1)
root.columnconfigure(0, weight=1)

textFrame = tk.Frame(root, bg='#101820', padx=5, pady=5)
controlFrame = tk.Frame(root, bg="#101010")#, width=2)
textFrame.grid(row=0, column=0, sticky="nsew")
controlFrame.grid(row=0, column=1, sticky="nsew")

searchLevel = 0  # 0 == no search, 1 = normal search, 2 = deep search

def on_value_change(*args):
    print("The selected option is:", float(temperature.get()))
    
# Create a style
style = ttk.Style()

# Modify the label style
style.configure("TLabel", background="grey", foreground="black", font=("Arial", 12))

# Modify the combobox style
style.configure("TCombobox", fieldbackground="grey", foreground="black", selectbackground="grey", font=("Arial", 12))
# Set the font of the dropdown list
style.configure("TCombobox", selectbackground='gray', fieldbackground='black', background='light gray')
style.map('TCombobox', fieldbackground=[('readonly','grey')])
style.configure("TCombobox", font=('Arial', 12))  # sets the font in the entry part
style.configure("TCombobox.Listbox", font=('Arial', 12), background='grey')  # sets the font in the dropdown list part



input_area = tk.Text(textFrame, height=80,bg='#101820', fg='AntiqueWhite1', font=("Bitstream Charter", 11), wrap=tk.WORD)
input_area.grid(row=0, column=0, sticky='nsew')
input_area.pack(fill=tk.BOTH, expand=1)
submit_button = tk.Button(controlFrame,  text="Submit", command=submit, font=("Arial", 12), bg='grey')
submit_button.pack(side='top', fill='x')

clear_button = tk.Button(controlFrame,  text="Clear", command=clear, font=("Arial",12), bg='grey')
clear_button.pack(side='top', fill='x')

temperature = tk.StringVar()
temperature.set('.1')
temperature.trace("w", on_value_change)
temp_label = ttk.Label(controlFrame, text="Temperature", style='TLabel')
temp_label.pack(side='top', fill='x')
temp_button = ttk.Combobox(controlFrame, textvariable=temperature, values = [.01,.1,.2,.4,.5,.7,.9,1.0], style='TCombobox')
temp_button.pack(side='top', fill='x')

top_p = tk.StringVar()
top_p.set('1.0')
top_p_label = ttk.Label(controlFrame, text="Top_p", style='TLabel')
top_p_label.pack(side='top', fill='x')
top_p_button = ttk.Combobox(controlFrame, textvariable=top_p, values = [.1,.2,.3,.4,.5,.6,.7,.8,.9,1.0], style='TCombobox')
top_p_button.pack(side='top', fill='x')

max_tkns = tk.StringVar()
max_tkns.set('250')
max_tok_label = ttk.Label(controlFrame, text="MaxTokens", style='TLabel')
max_tok_label.pack(side='top', fill='x')
max_tok_button = ttk.Combobox(controlFrame, textvariable=max_tkns, values = [50,100,250,500,1000,1500,2000,4000,8000], style='TCombobox')
max_tok_button.pack(side='top', fill='x')

prompt_button = tk.Button(controlFrame, text="setPrompt", command=setPrompt, font=("Arial", 12), bg='grey')
prompt_button.pack(side='top', fill='x')

#model_button = tk.Button(controlFrame, text="gpt-3.5-turbo", command=set_model, font=("Arial", 12), bg='grey')
#model_button.pack(side='top', fill='x')

root.rowconfigure(0, weight=1)
root.columnconfigure(0, weight=1)


root.mainloop()

"""
Find an arithmetic expression over the integers 2, 3, and 6 that evaluates to 12. You must use each integer exactly once. Use the following approach. Create a trial expression. Evaluate the results. If it is false, build hypothesis for generating a better expression, then generate a new trial expression and evaluate again. Do this until the correct result is found. If the result is 12 then the process is over.
"""

"""
Human: you are an instruction-following AI. Follow these instructions: Find an arithmetic expression over the integers 2, 3, and 6 that evaluates to 12. You must use each integer exactly once. Use the following approach.
Loop:
  Generate a trial expression using previously generated hypotheses, if any
  Evaluate the trial expression. 
  Test if the evaluation equals 12?
  If not, build a hypothesis to obtain higher success rate in expression generation.
Untill Test is True.
"""



from peft import PeftModel    
from transformers import AutoModelForCausalLM, AutoTokenizer, AutoModel
import transformers
import accelerate
import torch
from accelerate import load_checkpoint_and_dispatch
import datetime
import os
import traceback
from alphawave_pyexts import serverUtils as sv

# Load the model.
print('devices', torch.cuda.device_count(), torch.cuda.current_device())
model_name = "THUDM/chatglm2-6b"
print(f"Loading {model_name}")

tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)

model = AutoModel.from_pretrained(
        model_name,
        torch_dtype=torch.float16,
        device_map="auto",
        trust_remote_code=True
    )
model.tie_weights()
pipeline = transformers.pipeline(
    "text-generation",
    model=model,
    tokenizer=tokenizer,
    trust_remote_code=True,
    device_map="auto",
)


print(f"server started {model_name} into memory, use chatglm2 or maybe claude")

if __name__ == '__main__':
    #sv.server(tokenizer=tokenizer, pipeline=pipeline, stop_str=['Human:', 'Assistant:'])
    sv.server(model=model, tokenizer=tokenizer, stop_str=['###'])#, 'Human:'])#, 'Assistant:'])
